%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A two phase dual based Proximal ALM method for solving QP problems
%% Primal form: min_x 0.5*<x, Qx> + <c, x> s.t. Ax = b, x \in C:= {x | L <= x <= U}.
%% Dual form: max_{w, y, z} -0.5*<w, Qw> + <b, y> - delta_C^*(-z) s.t. -Qw + z + Aty = c, w \in Range(Q), 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INPUTS
%% (1) Qinput: 
%%         Qinput.Qmat matrix representation of Q; 
%%         Qinput.Qmap function handle of Q*x
%%         Note: if Qmat is not provided do not specify Qinput.Qmat, i.e., only input Qinput.Qmap 
%% (2) Initial points: x0, z0, w0, y0 are optional
%% (3) options: options for QPPAL
%%              options.stopTol: stopping tolerance in terms of KKT residuals (default: 1e-6)
%%              options.maxIterPhaseOne: maximum iteration of phase one (default: 1e3)
%%              options.maxIterPhaseTwo: maximum iteration of phase two (default: 1e3)
%%              options.sigma: initial sigma (default 1e0)
%%              options.sigmaPhaseTwo: if provided then use this value as inital sigma in phase two, otherwise use the one from phase one 
%%              options.kappa: inital value of kappa in phase two (default: 1e-4)
%%              options.iterComputeQx: compute Qx in phase one every iterComputeQx iterations (default 1)
%%              options.useSGS: whether to use sGS update in phase one (default: 1)
%%              options.AATMethod: method of AAT solver: direct (chol) or inexact (pcg)
%%              options.QMethod: method of (I+sig*Q) solver: direct (chol) or inexact (pcg)
%%              options.SSNMethod: method of SSN linear system solver: chol, lu, or bicgstab
%%              options.isDiagonalQ: whether Q is diagonal (default: 0), no need to specified
%%              options.printYes: whether to print run history in matlab window (default: 1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% OUTPUT
%% res: 
%%     res.pobj: primal objective value, 0.5*<x, Qx> + <c, x>
%%     res.dobj: dual objective value, -0.5*<w, Qw> + <b, y> - delta_C^*(-z)
%%     res.etag: relative gap: |pobj-dobj| / (1 + |pobj| + |dobj|)
%%     res.etap: primal infeasibility: ||Ax - b|| / (1 + ||b||) 
%%     res.etad: dual infeasibility: ||-Qw + z + Aty - c|| / (1 + ||c||)
%%     res.estQ: error between Qx and Qw: ||Qw - Qx|| / (1 + ||Qx|| + ||Qw||)
%%     res.etac: complementary error: ||x - Proj_C(x - z)|| / (1 + ||x|| + ||z||)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright @ 2021
%% Ling Liang, Xudong Li, Defeng Sun and Kim-Chuan Toh
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function res = QPPAL_main(Qinput, A, b, c, L, U, options, x0, z0, w0, y0)
    tstart = clock;
    rng('default');
    n = length(c);
    m = length(b);
    fprintf('\n *******************************************************');
    fprintf('******************************************');
    fprintf('\n QPPAL: A two phase algorithm for solving')
    if ~isempty(Qinput)
        fprintf(' convex QP');
    else
        fprintf(' LP');
    end
    fprintf('\n *******************************************************');
    fprintf('******************************************');
    %% initial points
    if nargin < 8
        x = zeros(n, 1);
        z = zeros(n, 1);
        w = zeros(n, 1);
        Qx = zeros(n, 1);
        Qw = zeros(n, 1);
        if m > 0
            y = zeros(m, 1);
        else 
            y = 0;
        end
    else
        %% check size
        if length(x0) ~= n || length(z0) ~= n || length(w0) ~= n || length(y0) ~= m
            error('Initial points provided have invalid sizes, please check again.');
        end
        x = x0;
        z = z0;
        w = w0;
        y = y0;
        Qx = Qinput.Qmap(x);
        Qw = Qinput.Qmap(w);
    end

    %% prepocessing
    [prob, sol, scl, AATSolve, QSolve, options, param] = ...
        prepocessingQP(Qinput, A, b, c, L, U, options, x, z, w, y, Qx, Qw);
    info.timePrepocessing = etime(clock, tstart);

    %% print header
    if options.printYes 
        fprintf('\n number of variables = %d,\n number of linear equality constraints = %d', prob.n, prob.m);
        fprintf('\n normb = %2.1e, normc = %2.1e', prob.normborg, prob.normcorg);
        fprintf('\n time for prepocessing = %2.1f', info.timePrepocessing);
        fprintf('\n *******************************************************');
        fprintf('******************************************');
    end

    runPhaseOne = options.maxIterPhaseOne > 0;
    ws = [];
    %% call QPPAL-Phase-One
    if runPhaseOne
        fprintf('\n Phase-I: An sCB-based ADMM\n');
        fprintf('\n  iter|  primal obj      dual obj        relgap | etap    etad    etaQ    etaC   | sigma   kapaa');
        tic;
        [prob, sol, scl, QSolve, param, infoPhaseOne] = QPPAL_Phase_One(prob, sol, scl, AATSolve, QSolve, options, param);
        timePhaseOne = toc;
        ws = infoPhaseOne.workSpace;
        fprintf('\n *******************************************************');
        fprintf('******************************************');
    end

    runPhaseTwo = options.twoPhase;
    %% call QPPAL-Phase-Two
    if runPhaseTwo
        fprintf('\n Phase-II: A proximal ALM\n');
        fprintf('\n  iter|  primal obj      dual boj        realgap| etap    etad    etaQ    etaC   | sigma   tau ');
        tic;
        [sol, infoPhaseTwo] = QPPAL_Phase_Two(prob, sol, scl, AATSolve, QSolve, ws, options, param);
        timePhaseTwo = toc;
        ws = infoPhaseTwo.workSpace;
        fprintf('\n *******************************************************');
        fprintf('******************************************');
    end

    %% recover solution and info
    if runPhaseOne
        res.infoPhaseOne = infoPhaseOne;
        res.timePhaseOne = timePhaseOne;
        res.iterPhaseOne = infoPhaseOne.iter;
    else
        res.infoPhaseOne = [];
        res.timePhaseOne = 0;
        res.iterPhaseOne = 0;
    end
    if runPhaseTwo
        res.infoPhaseTwo = infoPhaseTwo;
        res.timePhaseTwo = timePhaseTwo;
        res.iterPhaseTwo = infoPhaseTwo.iter;
    else
        res.infoPhaseTwo = [];
        res.timePhaseTwo = 0;
        res.iterPhaseTwo = 0;
    end
    res.x = (sol.x .* scl.cA) * scl.bscale;
    res.z = (sol.z ./ scl.cA) * scl.cscale;
    if prob.existA
        res.y = (sol.y .* scl.rA) * scl.cscale;
    end
    res.w = (sol.w .* scl.cA) * scl.bscale;
    res.pobj = ws.primobjorg;
    res.dobj = ws.dualobjorg;
    res.etag = ws.relgaporg;
    res.etap = ws.primfeasorg;
    res.etad = ws.dualfeasorg;
    res.etaq = ws.relerrQorg;
    res.etac = ws.relerrCorg;
    res.totalTime = etime(clock, tstart);

    %% print footer
    fprintf('\n Total time = %5.2f', res.totalTime);
    if runPhaseOne
        fprintf('\n Phase One: Time = %5.2f, Iter = %d', timePhaseOne, infoPhaseOne.iter);
    end
    if runPhaseTwo
        fprintf('\n Phase Two: Time = %5.2f, Iter = %d', timePhaseTwo, infoPhaseTwo.iter);
    end
    fprintf('\n primal objective = %+ 9.8e\n dual objective   = %+ 9.8e', ws.primobjorg, ws.dualobjorg);
    fprintf('\n etag = %+ 2.1e', ws.relgaporg);
    fprintf('\n etap = %+ 2.1e\n etad = %+ 2.1e\n etaQ = %+ 2.1e\n etaC = %+ 2.1e', ...
        ws.primfeasorg, ws.dualfeasorg, ws.relerrQorg, ws.relerrCorg);
    fprintf('\n *******************************************************');
    fprintf('******************************************\n');
end
