function [prob, sol, scl, AATSolve, QSolve, options, param] = prepocessingQP(Q, A, b, c, L, U, options, x, z, w, y, Qx, Qw)
    %% compute sizes
    m = length(b);
    n = length(c);
    
    %% check c
    if isempty(c)
        error('c is empty, please input an all-zero vector of proper size.');
    end
    if issparse(c)
        c = full(c);
    end
    
    %% check bounds
    if isempty(L)
        L = -1e20*ones(n, 1);
    end
    if isempty(U)
        U = 1e20*ones(n, 1); 
    end
    if issparse(L)
        L = full(L);
    end
    if issparse(U)
        U = full(U);
    end
    prob.Lorg = L;
    prob.Uorg = U;
    
    %% check Q
    existQ = ~isempty(Q);
    if ~existQ
        Q.Qmat = sparse(n, n);
        Q.Qmap = @(x) zeros(n, 1);
    end
    if isfield(Q, 'Qmat')
        if size(Q.Qmat, 1) ~= size(Q.Qmat, 2)
            error('Matrix representation of Q is provided, but it is not a square matrix.')
        end
%         if norm(Q.Qmat - Q.Qmat', 'fro') > 1e-14
%             error('Matrix representation of Q is provided, but it is not a symmetric matrix.')
%         end
        if size(Q.Qmat, 1) ~= n
            error('Matrix representation of Q is provided, sizes of Q and c are not consistent.')
        end
        fprintf('\n Matrix representation of Q is provided.');
        if ~isfield(Q, 'Qmap')
            Q.Qmap = @(x) Q.Qmat * x;
        end
    else
        x = randn(n, 1);
        try
            Qx = Q.Qmap(x);
            if length(Qx) ~= n
                error('Matrix representation of Q is not provided, but size of Q and c are not consistent.')
            end
        catch
            error('Matrix representation of Q is not provided, but Q as a linear mapping is not well-defined.')
        end
        fprintf('\n Matrix representation of Q is not provided.');
    end
    
    %% check A
    existA = ~isempty(A);
    if existA && isempty(b)
        error('A exists but b is empty, please check the constraint data.');
    end
    if ~existA && ~isempty(b)
        error('A is empty but b is not empty, please check the constraint data.');
    end
    if size(A, 1) ~= m && existA
        error('Number of rows of A is not equal to length of b.');
    end
    if size(A, 2) ~= n && existA
        error('Number of cols of A is not equal to length of c.');
    end
    if existA && issparse(b)
        b = full(b); 
    end
    
    %% parse options
    if ~isfield(options, 'stopTol')
        options.stopTol = 1e-6;
    end
    if ~isfield(options, 'stopTolPhaseOne')
        options.stopTolPhaseOne = 1e-4;
    end
    if ~isfield(options, 'maxIterPhaseOne')
        options.maxIterPhaseOne = 1000;
    end
    if ~isfield(options, 'maxIterPhaseTwo')
        options.maxIterPhaseTwo = 1000;
    end
    if ~isfield(options, 'twoPhase')
        options.twoPhase = 1;
    end
    if ~isfield(options, 'maxIterStep')
        options.maxIterStep = 40;
    end
    if ~isfield(options, 'maxIterSSN')
        options.maxIterSSN = 100;
    end
    if ~isfield(options, 'sigma')
        options.sigma = 1e0;
    end
    if ~isfield(options, 'gamma')
        options.gamma = 1.95;
    end
    if ~isfield(options, 'kappa')
        options.kappa = 1e-2;
    end
    if ~isfield(options,'QMethod')
        options.QMethod = 'direct';
        if ~isfield(Q, 'Qmat')
            options.QMethod = 'inexact';
        end
    end
    if ~isfield(options,'AATMethod')
        options.AATMethod = 'direct';
    end
    if ~isfield(options,'SSNMethod')
        if n < 9000 || (nnz(A)/numel(A) < 1e-3 && isfield(Q, 'Qmat') && nnz(Q.Qmat) / numel(Q.Qmat) < 1e-3)
            options.SSNMethod = 'chol';
        else
            options.SSNMethod = 'lu';
        end
        if ~isfield(Q, 'Qmat')
            options.SSNMethod = 'bicgstab';
        end
    end
    if ~isfield(options, 'scaleA')
        if min(norm(b), norm(c)) < 1e-10 && false
            options.scaleA = 0;
        else
            options.scaleA = 1;
        end
    end
    if ~isfield(options, 'scaleData')
        options.scaleData = 1;
    end
    if ~isfield(options, 'rescale')
        options.rescale = 1;
    end
    if ~isfield(options, 'useSGS')
        options.useSGS = 1;
    end
    if ~isfield(options, 'sigmaFix')
        options.sigmaFix = 0;
    end
    if ~isfield(options, 'sigmaMax')
        options.sigmaMax = 5e6;
    end
    if ~isfield(options, 'sigmaMin')
        options.sigmaMin = 1e-4;
    end
    if ~isfield(options, 'sigmaScale')
        options.sigmaScale = 1.35;
    end
    if ~isfield(options, 'printYes')
        options.printYes = 1;
    end
    if ~isfield(options, 'projPhaseOne')
        options.projPhaseOne = 0;
    end
    if ~isfield(options, 'projPhaseTwo')
        options.projPhaseTwo = 1;
    end
    if ~isfield(options, 'iterComputeQx')
        options.iterComputeQx = 1;
    end
    if ~isfield(options, 'useEigs')
        options.useEigs = 0;
    end
    if ~isfield(options, 'isDiagonalQ')
        diagCheckQ = 1;
    else
        diagCheckQ = 0;
    end
    if ~isfield(options, 'sigmaUpdateIter')
        options.sigmaUpdateIter = 50;
    end
    if ~isfield(options, 'relgapConst')
        options.relgapConst = 1e0;
    end
        

    %% compute norms
    normborg = norm(b);
    normcorg = norm(c);
    if options.useEigs
        rQ = 1;
        eigsopt.v0 = randn(n, 1);
        eigsopt.tol = 1e-6;
        eigsopt.maxit = 50;
        normQorg = eigs(Q.Qmap, n, rQ, 'LA', eigsopt);
    else
        normQorg = inf;
    end

    %% scaling A
    if existA
        if options.scaleA == 1
            maxIterRuizEqb = 50;
            [rA, cA, A] = RuizEquilibration(A, maxIterRuizEqb);
        elseif options.scaleA == 2
            rA = 1 ./ min(1e3, max(1e-3, sqrt(sum(A.*A, 2))));
            cA = 1 ./ min(1e3, max(1e-3, sqrt(sum(A.*A, 1)')));
            A = spdiags(rA, 0, lb, lb) * (A * spdiags(cA, 0, lc, lc));
        else
            rA = ones(m, 1);
            cA = ones(n, 1);
        end
        b = b .* rA;
        c = c .* cA;
        x = x ./ cA;
        y = y ./ rA;
        z = z .* cA;
        w = w ./ cA;
        Qx = Qx .* cA;
        Qw = Qw .* cA;
        L = L ./ cA;
        U = U ./ cA;
        if existQ
            if isfield(Q, 'Qmat')
                CA = spdiags(cA, 0, n, n);
                Q.Qmat = CA * (Q.Qmat * CA);
            end
            if isfield(Q, 'Qmap')
                Q.Qmap = @(x) (Q.Qmap(x .* cA)) .* cA;
            end
        end
    end

    %% scale data
    if options.scaleData == 1
        bscale = max(1, norm(b));
        cscale = max(1, norm(c));
    else
        bscale = 1;
        cscale = 1;
    end
    b = b / bscale;
    c = c / cscale;
    L = L / bscale;
    U = U / bscale;
    x = x / bscale;
    z = z / cscale;
    w = w / bscale;
    y = y / cscale;
    Qx = Qx / cscale;
    Qw = Qw / cscale;
    if existQ
        bcratio = bscale / cscale;
        if isfield(Q, 'Qmat')
            Q.Qmat = Q.Qmat * bcratio;
        end
        if isfield(Q, 'Qmap')
            Q.Qmap = @(x) Q.Qmap(x) * bcratio;
        end
    end 

    %% AATSolver
    if existA
        if strcmp(options.AATMethod, 'direct')
            fprintf('\n Using direct method to handle AAT');
            [LAAT, AAT] = mycholAAt(A);
            AATSolve.AAT = AAT;
            AATSolve.diagAAT = diag(AAT);
            AATSolve.solveFun = @(x) mylinsysolve(LAAT, x);
            AATSolve.L = LAAT;
        elseif strcmp(options.AATMethod, 'inexact')
            fprintf('\n Using PCG method to handle AAT');
            parA.precond = 1;
            if parA.precond == 1
                parA.invdiagM = 1 ./ max(1e-3, min(1e3, full(sum(A.*A, 2))));
            end
            At = A';
            AATmap = @(x) A*(At*x); %% AAT may be dense, compute separately
            AATSolve.solveFun = @(rhs, tol) mypsqmr(AATmap, rhs, parA, tol, 5000);
            AATSolve.parA = parA;
        end
    else
        AATSolve = [];
    end

    %% QSolver
    if existQ
        sigma = options.sigma;
        %% check if Q.Qmat is diagonal
        if isfield(Q, 'Qmat')
            diagQ = diag(Q.Qmat);
            QSolve.diagQ = diagQ;
            if diagCheckQ
                if norm(Q.Qmat - spdiags(diagQ, 0, n, n), 'fro') < 1e-15
                    fprintf('\n Q is diagonal, inv(I+sigma*Q) has analytical expression')
                    QSolve.solveFun = @(rhs) rhs ./ (1 + sigma*diagQ);
                    options.QMethod = 'diagonal';
                end
            end
        end
        if strcmp(options.QMethod, 'direct')
            fprintf('\n Using chol to calculate inv(I+sigma*Q)');
            LQ = mycholQ(Q.Qmat, n, sigma);
            QSolve.L = LQ;
            QSolve.solveFun = @(rhs) mylinsysolve(LQ, rhs);
        elseif strcmp(options.QMethod, 'inexact')
            fprintf('\n Using PCG to calculate inv(I+sigma*Q)');
            parQ.sig = sigma;
            if options.useEigs
                rQ = min(10, n);
                [VQ, dQ] = eigs(Q.Qmap, n, rQ, 'LA', eigsopt);
                dQ = diag(dQ);
                [dQ, idxSortQ] = sort(dQ, 'descend');
                VQ = VQ(:, idxSortQ);
                idxPosQ = find(dQ > 1e-16);
                if ~isempty(idxPosQ)
                    dQ = dQ(idxPosQ);
                    VQ = VQ(:, idxPosQ);
                    parQ.precond = 2;
                    parQ.V = VQ;
                    parQ.d = dQ;
                else
                    if isfield(Q, 'Qmat')
                        parQ.precond = 1;
                        parQ.diagQ = diagQ;
                        parQ.invdiagM = 1 ./ (1 + sigma*diagQ);
                    else
                        parQ.precond = 0;
                    end
                end
            else
                if isfield(Q, 'Qmat')
                    parQ.precond = 1;
                    parQ.diagQ = diagQ;
                    parQ.invdiagM = 1 ./ (1 + sigma*diagQ);
                else
                    parQ.precond = 0;
                end
            end
            QSolve.parQ = parQ;
        end
    end

    %% collect data      
    prob.Q = Q;
    prob.A = A;
    prob.At = A';
    prob.b = b;
    prob.c = c;  
    prob.L = L;
    prob.U = U;
    prob.m = m;
    prob.n = n;
    prob.normborg = normborg;
    prob.normcorg = normcorg;
    prob.normQorg = normQorg;
    prob.normb = norm(b);
    prob.normc = norm(c);
    prob.existQ = existQ;
    prob.existA = existA;

    sol.x = x;
    sol.z = z;
    sol.w = w;
    sol.y = y;
    sol.Qx = Qx;
    sol.Qw = Qw;

    if existA
        scl.cA = cA;
        scl.rA = rA;
    else
        scl.cA = 1;
        scl.rA = 1;
    end
    scl.bscale = bscale;
    scl.cscale = cscale;
    scl.objscale = bscale * cscale;

    param.sigma = options.sigma;
    param.gamma = options.gamma;
    param.kappa = options.kappa;
end %% end of validation